//
// Eight+one band binary tree predictive encoder
//
// Version 4. Includes new tree structure
//
// Copyright (c) John A Robinson. 1994, 1995, 1997
//
#include <stdlib.h>
#include <math.h>
#include <string.h>
#include "compact.hpp"

void rgb_ycc(unsigned char ***components,int ymax,int xmax, int smooth);
int makescratch(const int& xmax, const int& ymax);
int deletescratch(const int& ymax);
void btpc_enc_plane(unsigned char **in, const int& xmax, const int& ymax,
const int& fromlevel, const int& tolevel, const int *spacing, 
compactor *codeout, const int& clipval);

//
// Calculation of quantization levels from quality parameter
//
int calcq(int maxval, int quality, int *qvals, int numbands) {
	if ((quality > 100) || (quality < 0))
		return -1;
	int topval = (100-quality)/2;
	// topval will be the quantizer spacing on the finest band
	topval *= maxval;
	topval /= 255;
	if (!topval) {	// Quality == 100 or scaling makes it so
		for (int i = 0; i < numbands; i++)
			*qvals++ = 1;
		return 1;
		}
	float qratio;
	if (topval >= 40)
		{
		topval += (topval-25) + 3*(topval-40);
		qratio = 0.75;
		}
	else if (topval > 25)
		{
		topval += (topval-25);
		qratio = 0.75;
		}
	else
		qratio = 0.8;
	qvals[numbands-1] = topval;
	for (int i = numbands-2; i >= 0; i--)
		qvals[i] = (int)((((float)qvals[i+1])*qratio)+0.99999);
	return 1;
	}



void pgmword(FILE *in, char *cbuf)
// The PGM man page says you have to be as lenient as possible about pgm
// files, so this is my general-purpose lenient pgm word reader
// Read next PGM word, ignoring comments and whitespace
	{
	char c;
	while(1) {
	    do 
		c = fgetc(in);
	    while(c == ' ' || c == '\t' || c == '\n' || c == '\r');
	    if (c == '#') {
		do
			c = fgetc(in);
		while (c != '\n');
		}
	    else break;
	    }
	*cbuf++ = c;
	c = fgetc(in);
	while (c != ' ' && c != '\t' && c != '\n' && c != '\r' && c != '#')
	    {
	    *cbuf++ = c;
	    c = fgetc(in);
	    }
	fseek(in, -1, SEEK_CUR);
	*cbuf = '\0';
	}

int readpnmandpad(const char *pgmname, int& rows, int& cols, 
	int& maxval, unsigned char ***pbuf, int numlevels)
// Read picture in "P5" or "P6" format into arrays. Pad the picture size up to
// the next multiple of blocksize.
	{
	int i,j;
	char cbuf[32];
	unsigned char c;
	int retval = 0;
	FILE *in=fopen(pgmname, "rb");
	if (!in) {
		fprintf(stderr, "Cannot open %s\n", pgmname);
		return 0;
		}
	cbuf[0] = fgetc(in);
	cbuf[1] = fgetc(in);
	cbuf[2] = 0;
	if (!strcmp(cbuf,"P5")) 
		// PGM image
		retval = 1;
	else if (!strcmp(cbuf,"P6"))
		// PPM image
		retval = 3;
	else {
		fprintf(stderr,"Sorry. This version of cbtpc only works with ");
		fprintf(stderr,"raw format (P5 or P6) PGM or PPM files.\n");
		fprintf(stderr,"%s is not a valid raw format file\n",pgmname);
		return 0;
		}
	pgmword(in, cbuf);
	cols = atoi(cbuf);
	pgmword(in,cbuf);
	rows = atoi(cbuf);
	pgmword(in,cbuf);
	maxval = atoi(cbuf);
	if (maxval > 255)	// Shouldn't happen in a valid raw pgm file
		maxval = 255;	// ... but just in case it does
	// absorb final whitespace char before start of data
	c = fgetc(in);
	// Make ymax equal to next highest multiple of blocksize
	int ymax = ((rows+(1<<numlevels)-1)>>numlevels)<<numlevels;
	// Make xmax equal to next highest multiple of 2*blocksize
	// Do this so that points can be processed in pairs in binary tree
	int xmax = ((cols+(1<<(numlevels+1))-1)>>(numlevels+1))<<(numlevels+1);
	if (retval == 1)	// PGM straightforward
	    {
	    unsigned char **buf;
	    *pbuf = buf = new unsigned char *[ymax];
	    for (i = 0; i < rows; i++) {
		buf[i] = new unsigned char[xmax];
		fread(buf[i], 1, cols, in);
		for (j = cols; j < xmax; j++)	// Pad image by
			buf[i][j] = buf[i][cols-1];	// copying last pel
		}
	    // Vertical padding
	    for (i = rows; i < ymax; i++) {
		buf[i] = new unsigned char[xmax];	// Pad image by
		memcpy(buf[i], buf[rows-1], xmax);	// copying last line
		}
	    }
	else			// PPM
	    {
	    unsigned char **bufr, **bufg, **bufb;
	    unsigned char *tempbuf = new unsigned char[3*xmax];
	    unsigned char *pr, *pg, *pb, *ptemp;
	    *pbuf++ = bufg = new unsigned char *[ymax];
	    *pbuf++ = bufr = new unsigned char *[ymax];
	    *pbuf = bufb = new unsigned char *[ymax];
	    // This loop incorporates vertical padding
	    for (i = 0; i < ymax; i++){
		pg = bufg[i] = new unsigned char[xmax];
		pr = bufr[i] = new unsigned char[xmax];
		pb = bufb[i] = new unsigned char[xmax];
		if (i < rows)
			fread(tempbuf,1,cols*3,in);
		ptemp = tempbuf;
		for (j = 0; j < cols; j++) {
			*pr++ = *ptemp++;
			*pg++ = *ptemp++;
			*pb++ = *ptemp++;
			}
		for (j = cols; j < xmax; j++) {	// Pad
			*pr++ = bufr[i][cols-1];
			*pg++ = bufg[i][cols-1];
			*pb++ = bufb[i][cols-1];
			}
		}
	    delete [] tempbuf;
	    }
	fclose(in);
	return retval;
	}


int main(int argc, const char *argv[])
	{
	const int numlevels = 4;  // Number of levels in tree
	if ((argc != 3)&&(argc != 4))
		{
		fprintf(stderr,"Usage: cbtpc pgm_inpic code_file [quality]\n");
		return -1;
		}
	unsigned char **components[3];
	unsigned char **in;
	int truerows, truecols;
	int maxval;
	int pnmtype;
	int i;
	int compnum;
	char coltype = 'G';	// GRB or monochrome
	char uvfilter = '0';	// No downsampling
	// Get picture into unsigned char arrays
	pnmtype = readpnmandpad(argv[1],truerows,truecols,maxval,components,
		numlevels);
	if (!pnmtype)
		return -1;
	// Make ymax equal to next highest multiple of blocksize
	int ymax = ((truerows+(1<<numlevels)-1)>>numlevels)<<numlevels;
	// Make xmax equal to next highest multiple of 2*blocksize
	// Do this so that points can be processed in pairs in binary tree
	int xmax = ((truecols+(1<<(numlevels+1))-1)>>(numlevels+1))<<(numlevels+1);
	int spacing[65];	// Quantizer spacings
	int quality = 75;
	if (argc == 4)
		quality = atoi(argv[3]);
	if (calcq(maxval, quality, spacing+1, numlevels*2) < 0) {
		fprintf(stderr,"Quality parameter must be between 0 and 100\n");
		return -1;
		}
	if ((pnmtype == 3)&&(spacing[numlevels*2] >= 2))
		{
		coltype = 'Y';
		uvfilter = 'B';	// Box filter
		rgb_ycc(components,ymax,xmax,1);
		}
	FILE *cfp = fopen(argv[2],"wb");
	if (cfp == 0)
		{
		fprintf(stderr,"Can't open %s\n", argv[2]);
		return -1;
		}
	// Write header
	if (pnmtype == 1)	
		fprintf(cfp,"btpc 4.b1%c%cG0\n%d\n%d\n", 
			numlevels+'0', numcoders+'0', truerows, truecols);
	else
		fprintf(cfp,"btpc 4.b3%c%c%c%c\n%d\n%d\n", 
			numlevels+'0', numcoders+'0', coltype, uvfilter, 
			truerows, truecols);
	// Output all the quantizer spacing parameters
	for (i = 1; i <=numlevels*2; i++)
		fputc((unsigned char) spacing[i], cfp);
	// And the maxval parameter
	fputc((unsigned char) maxval, cfp);
	// Set up the huffman coders
	cfileio fio(cfp);
	compactor codeout[numcoders];
	for (i = 0; i < numcoders; i++)
		codeout[i].attach(&fio);
	// Encode each plane of the picture
	makescratch(xmax, ymax);
	for (compnum = 0; compnum < pnmtype; compnum++) {
	    int tolevel = (compnum && (uvfilter != '0'));
	    int clipval = tolevel ? 255 : maxval;
	    btpc_enc_plane(components[compnum],xmax,ymax,numlevels,
		tolevel, spacing, codeout, clipval);
	    }
	fio.flush();
	fclose(cfp);
	// And finally, clean up
	for (compnum = 0; compnum < pnmtype; compnum++) {
	    in = components[compnum];
	    for (i = 0; i < ymax; i++)
		delete [] in[i];
	    delete [] in;
	    }
	deletescratch(ymax);
	return 1;
	}

